/***************************************************************************
             centerline.c  -  convert pixels to centerlines
                             -------------------
    begin                : Thu Sep 5 2002
    copyright            : (C) 2002 by Adam Tofilski
    email                : rotofils@cyf-kr.edu.pl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xstd.h"
#include "pxl-outline.h"

#define MOD(x) ((x) < 0 ? -x : x)

void
del_short_otl(pixel_outline_list_type *outlines, unsigned otl_nr,
              unsigned too_short);


/* Works only with 8 bit per pixel bitmaps. */
unsigned char
GetColor(at_bitmap_type bitmap, unsigned int row, unsigned int col)
{
  unsigned pixel=row*bitmap.np*bitmap.width+col*bitmap.np;

  return bitmap.bitmap[pixel];
}

/* Works only with 8 bit per pixel bitmaps. */
void
SetColor(at_bitmap_type *bitmap, unsigned int row, unsigned int col,
         unsigned char color)
{
  unsigned pixel=row*bitmap->np*bitmap->width+col*bitmap->np;

  bitmap->bitmap[pixel]=color;
}


/* Count all neighbours of a pixel. */
unsigned
count_all_neighbours(at_bitmap_type bitmap, unsigned row, unsigned col)
{
  unsigned neighbors=0;

  if(GetColor(bitmap,row+1, col+1)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row+1, col)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row+1, col-1)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row, col+1)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row, col-1)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row-1, col+1)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row-1, col)!=0xFF)
    neighbors++;
  if(GetColor(bitmap,row-1, col-1)!=0xFF)
    neighbors++;

  return neighbors;
}


/* Count neighbours of a pixel, but not in corners with 2 neighbours. */
unsigned
count_neighbours(at_bitmap_type bitmap, unsigned row, unsigned col)
{
  unsigned neighbors=0, NewRow, NewCol;
/*  unsigned color=GetColor(bitmap, row, col);*/

  NewRow=row+1;                                    /*      xxx    */
  if(GetColor(bitmap,NewRow, col)!=0xFF)           /*      x*x    */
    neighbors++;                                   /*      x0x    */
  else
    {
      if(GetColor(bitmap,NewRow, col+1)!=0xFF)     /*      xxx    */
        if(GetColor(bitmap,row, col+1)==0xFF)      /*      x*1    */
          neighbors++;                             /*      x10    */

      if(GetColor(bitmap,NewRow, col-1)!=0xFF)     /*      xxx    */
        if(GetColor(bitmap,row, col-1)==0xFF)      /*      1*x    */
          neighbors++;                             /*      01x    */
    }
  NewRow=row-1;                                    /*      x0x    */
  if(GetColor(bitmap,NewRow, col)!=0xFF)           /*      x*x    */
    neighbors++;                                   /*      xxx    */
  else
    {
      if(GetColor(bitmap,NewRow, col+1)!=0xFF)     /*      x10    */
        if(GetColor(bitmap,row, col+1)==0xFF)      /*      x*1    */
          neighbors++;                             /*      xxx    */

      if(GetColor(bitmap,NewRow, col-1)!=0xFF)     /*      01x    */
        if(GetColor(bitmap,row, col-1)==0xFF)      /*      1*x    */
          neighbors++;                             /*      xxx    */
    }

  NewCol=col+1;                                    /*      xxx    */
  if(GetColor(bitmap, row, NewCol)!=0xFF)          /*      x*0    */
    neighbors++;                                   /*      xxx    */

  NewCol=col-1;                                    /*      xxx    */
  if(GetColor(bitmap, row, NewCol)!=0xFF)          /*      0*x    */
    neighbors++;                                   /*      xxx    */

  return neighbors;
}


/* Find unmarked neighbours. If not found return false else return
   true and change row and col to point to found pixel. */
at_bool
FindUnmarkedPixel(at_bitmap_type bitmap, bitmap_type marked,
                  unsigned short *row, unsigned short *col, at_bool direction)
{
  unsigned NewRow, NewCol;
  int dir;

  if(direction)
    dir=1;
  else
    dir=-1;

  /* Don not change sequence of checking. First check horizontaly and verticaly
     later in corners. */
  NewCol=*col+dir;
  if((GetColor(bitmap, *row, NewCol)==0)
      && (GetColor(marked, *row, NewCol)==0xFF))
    {
      *col=NewCol;
      return true;
    }

  NewCol=*col-dir;
  if((GetColor(bitmap, *row, NewCol)==0)
      && (GetColor(marked, *row, NewCol)==0xFF))
    {
      *col=NewCol;
      return true;
    }

  NewRow=*row+dir;
  if((GetColor(bitmap, NewRow, *col)==0)
      && (GetColor(marked, NewRow, *col)==0xFF))
    {
      *row=NewRow;
      return true;
    }
  NewCol=*col+dir;
  if((GetColor(bitmap, NewRow, NewCol)==0)
      && (GetColor(marked, NewRow, NewCol)==0xFF))
    {
      *row=NewRow;
      *col=NewCol;
      return true;
    }
  NewCol=*col-dir;
  if((GetColor(bitmap, NewRow, NewCol)==0)
      && (GetColor(marked, NewRow, NewCol)==0xFF))
    {
      *row=NewRow;
      *col=NewCol;
      return true;
    }

  NewRow=*row-dir;
  if((GetColor(bitmap, NewRow, *col)==0)
      && (GetColor(marked, NewRow, *col)==0xFF))
    {
      *row=NewRow;
      return true;
    }
  NewCol=*col+dir;
  if((GetColor(bitmap, NewRow, NewCol)==0)
      && (GetColor(marked, NewRow, NewCol)==0xFF))
    {
      *row=NewRow;
      *col=NewCol;
      return true;
    }
  NewCol=*col-dir;
  if((GetColor(bitmap, NewRow, NewCol)==0)
      && (GetColor(marked, NewRow, NewCol)==0xFF))
    {
      *row=NewRow;
      *col=NewCol;
      return true;
    }

  return false;
}


/* Concatenate 2 pixel outlines. The two lists are assumed to have the
   same last pixel. */
/* o1------>T * T<------o2  */
void
concatTT(pixel_outline_type *o1, pixel_outline_type *o2)
{
  int src, dst;
  unsigned o1_length, o2_length;

  if (!o1 || !o2 || O_LENGTH(*o2) <= 1)
    return;

  o1_length = O_LENGTH(*o1);
  o2_length = O_LENGTH(*o2);
  O_LENGTH(*o1) += o2_length - 1;
  /* Resize o1 to the sum of the lengths of o1 and o2 minus one (because
     the two lists are assumed to share the same starting pixel). */
  XREALLOC(o1->data, O_LENGTH(*o1) * sizeof(at_coord));

  /* Prepend the contents of o2 (in reverse order) to o1. */
  for (src = o2_length - 2, dst = o1_length; src >= 0; src--, dst++)
    O_COORDINATE(*o1, dst) = O_COORDINATE(*o2, src);

  o1->to_junction = o2->from_junction;
  o2->from_junction = NULL;
  o2->to_junction = NULL;

  if (o2->data != NULL)
  {
    free (o2->data);
    o2->data = NULL;
  }
  o2->length=0;
}


/* Concatenate 2 pixel outlines. The last pixel of the first lists
   is assumed to be the first pixel of the second list. */
/* o1------>T * F>------o2  */
void
concatTF(pixel_outline_type *o1, pixel_outline_type *o2)
{
  unsigned src, dst;
  unsigned o1_length, o2_length;

  if (!o1 || !o2 || O_LENGTH(*o2) <= 1)
    return;

  o1_length = O_LENGTH(*o1);
  o2_length = O_LENGTH(*o2);
  O_LENGTH(*o1) += o2_length - 1;
  /* Resize o1 to the sum of the lengths of o1 and o2 minus one (because
     the two lists are assumed to share the same starting pixel). */
  XREALLOC(o1->data, O_LENGTH(*o1) * sizeof(at_coord));

  /* Prepend the contents of o2 to o1. */
  for (src = 1, dst = o1_length; src < o2_length; src++, dst++)
    O_COORDINATE(*o1, dst) = O_COORDINATE(*o2, src);

  o1->to_junction = o2->to_junction;
  o2->from_junction = NULL;
  o2->to_junction = NULL;

  if (o2->data != NULL)
  {
    free (o2->data);
    o2->data = NULL;
  }
  o2->length=0;
}


/* Concatenate 2 pixel outlines. The two lists are assumed to have the
   same first pixel. */
/* o1------<F * F>------o2  */
void
concatFF(pixel_outline_type *o1, pixel_outline_type *o2)
{
  concat_pixel_outline(o1, o2);

  o1->from_junction = o2->to_junction;
  o2->from_junction = NULL;
  o2->to_junction = NULL;

  if (o2->data != NULL)
  {
    free (o2->data);
    o2->data = NULL;
  }
  o2->length=0;
}

/* Find outlines conected to junction. */
pixel_outline_type *
find_otl(unsigned start, pixel_outline_list_type *outlines,
         at_coord junction, at_bool *to)
{
  unsigned row = junction.y;
  unsigned col = junction.x;
  unsigned otl_j;

  for(otl_j=start; otl_j<outlines->length; otl_j++)
    {
      if(outlines->data[otl_j].length)     //skip deleted
      {
        if(outlines->data[otl_j].to_junction)
          if((outlines->data[otl_j].to_junction->y == row)
             && (outlines->data[otl_j].to_junction->x == col))
          {
            *to = true;
            return &outlines->data[otl_j];
          }
        if(outlines->data[otl_j].from_junction)
          if((outlines->data[otl_j].from_junction->y == row)
             && (outlines->data[otl_j].from_junction->x == col))
          {
            *to = false;
            return &outlines->data[otl_j];
          }
      }
    }
  return NULL;
}



/* If outline is too short, delet it and concat neighbours. */
void
del_short_otl(pixel_outline_list_type *outlines, unsigned otl_nr,
              unsigned too_short)
{
  pixel_outline_type *otl = &(outlines->data[otl_nr]);

  if((otl->length < too_short)
      &&(otl->length!=0)               /* contated otl have length 0 */
      &&((otl->from_junction == NULL)  /* not conected on bot sides */
      ||(otl->to_junction == NULL))){

    LOG2("del: #%d, length: [%d]\n", otl_nr,
         outlines->data[otl_nr].length);
    otl->length=0;

    if((otl->from_junction == NULL)    /* not from any junction */
        && (otl->to_junction == NULL))
      return;
    else
    {
      pixel_outline_type *otl0 = NULL;
      pixel_outline_type *otl1 = NULL;
      pixel_outline_type *otl2 = NULL;
      at_bool to0=false;
      at_bool to1=false;
      at_bool to2=false;
      at_coord *junction;

      if(otl->to_junction)
        junction = otl->to_junction;
      else
        junction = otl->from_junction;

      otl0 = find_otl(0, outlines, *junction, &to0);
      if(otl0)
      {
        otl1 = find_otl(otl0->nr+1, outlines, *junction, &to1);
        if(otl1)
        {
          otl2 = find_otl(otl1->nr+1, outlines, *junction, &to2);
          if(otl2)
          {
              LOG("junction of 4 outlines\n");
              return;
          }
        }
        else
        {
          LOG("otl1 not found\n");
          return;
        }
      }
      else
      {
        LOG("otl0 not found\n");
        return;
      }
      if(to0){
        if(to1){
          LOG2("concat TT: #%d, #%d\n", otl0->nr, otl1->nr);
          concatTT(otl0, otl1);
          del_short_otl(outlines, otl0->nr, too_short);
        }
        else{
          LOG2("concat TF: #%d, #%d\n", otl0->nr, otl1->nr);
          concatTF(otl0, otl1);
          del_short_otl(outlines, otl0->nr, too_short);
        }
      }
      else
      {
        if(to1){
          LOG2("concat TF: #%d, #%d\n", otl1->nr, otl0->nr);
          concatTF(otl1, otl0);           /* arguments order important */
          del_short_otl(outlines, otl1->nr, too_short);
        }
        else{
          LOG2("concat FF: #%d, #%d\n", otl0->nr, otl1->nr);
          concatFF(otl0, otl1);
          del_short_otl(outlines, otl0->nr, too_short);
        }
      }

    }
  }
}

/*
Find outlines shorter than too_short, delet them and concat neighbours.
*/
void
find_short_otl (pixel_outline_list_type *outlines, unsigned too_short)
{
  unsigned otl_nr;

  for (otl_nr=2; otl_nr<outlines->length; otl_nr++)
    del_short_otl (outlines, otl_nr, too_short);
}


/*  Mark pixel to avoid revisiting: marked 0x00, unmarked 0xFF.  */
void
MarkPixel(at_bitmap_type *bitmap, unsigned row, unsigned col)
{
  SetColor(bitmap, row, col, 0); /* marked 0x00 unmarked 0xFF */
}


/* Unmark pixel at junction to allow revisiting: marked 0x00, unmarked 0xFF. */
void
UnmarkPixel(at_bitmap_type *bitmap, unsigned row, unsigned col)
{
  SetColor(bitmap, row, col, 0xFF);
}


at_bitmap_type
empty_bmp(unsigned width, unsigned height, unsigned np)
{
  unsigned int row, col;
  at_bitmap_type new_bmp;

  new_bmp.width=width;
  new_bmp.height=height;
  new_bmp.np=np;
  XMALLOC (new_bmp.bitmap, new_bmp.width * new_bmp.height
            * sizeof (unsigned char));

  for (row = 0; row < new_bmp.height; row++)
    for (col = 0; col < new_bmp.width; col++)
      SetColor(&new_bmp, row, col, 0xFF);
  return new_bmp;
}


/* from Autotrace pxl-outline.c  modified */
pixel_outline_type
find_one_centerline(bitmap_type bitmap, bitmap_type *marked,
                    unsigned short original_row, unsigned short original_col)
{

  pixel_outline_type otl = new_pixel_outline();
  unsigned short row = original_row, col = original_col;
  at_coord pos;

  otl.color = GET_COLOR(bitmap, row, col);

  do
    {
      pos.x = col; pos.y = bitmap.height - row;
      append_outline_pixel(&otl, pos);

      MarkPixel(marked, row, col);
      LOG2("(%d,%d)", pos.x, pos.y);

      if (count_neighbours(bitmap, row, col)>2)
        {
          if((row!=original_row)||(col!=original_col))
            {
              XMALLOC(otl.to_junction, sizeof(at_coord *));
              otl.to_junction->y=row;
              otl.to_junction->x=col;
              LOG("J");
              break;
            }
          else
            {
              XMALLOC(otl.from_junction, sizeof(at_coord *));
              otl.from_junction->y=row;
              otl.from_junction->x=col;
              LOG("J");
            }
        }
      }while(FindUnmarkedPixel(bitmap, *marked, &row, &col, true));

  if ((MOD(row-original_row)<2)&&(MOD(col-original_col)<2)&&(otl.length>3))
    otl.open = false;
  else
    otl.open = true;
  LOG("//");

  if((otl.from_junction==NULL)&&(otl.open))
    {
      pixel_outline_type otl_part = new_pixel_outline();

      row=original_row;
      col=original_col;

    do
      {
        pos.x = col; pos.y = bitmap.height - row;
        append_outline_pixel(&otl_part, pos);

        MarkPixel(marked, row, col);
        LOG2("(%d,%d)", pos.x, pos.y);

        if (count_neighbours(bitmap, row, col)>2)
          {
             XMALLOC(otl.from_junction, sizeof(at_coord *));
          otl.from_junction->y=row;
          otl.from_junction->x=col;
          LOG("J");
          break;
          }
      }while(FindUnmarkedPixel(bitmap, *marked, &row, &col, false));
      concat_pixel_outline(&otl, &otl_part);
  }

  if(otl.from_junction)
    UnmarkPixel(marked, otl.from_junction->y, otl.from_junction->x);
  if(otl.to_junction)
    UnmarkPixel(marked, otl.to_junction->y, otl.to_junction->x);

  return otl;
}

/* from Autotrace pxl-outline.c  modified */
#ifdef _EXPORTING
__declspec(dllexport) void
__stdcall find_centerline_pixels (pixel_outline_list_type *outline_list,
                                  bitmap_type bitmap, color_type bg_color,
                                  at_progress_func notify_progress, address progress_data,
                                  at_testcancel_func test_cancel, address testcancel_data)
#else
void
find_centerline_pixels(pixel_outline_list_type *outline_list,
                       bitmap_type bitmap, color_type bg_color,
                       at_progress_func notify_progress, at_address progress_data,
                       at_testcancel_func test_cancel, at_address testcancel_data)
#endif
{
  unsigned short row, col;
  bitmap_type marked = empty_bmp(BITMAP_WIDTH(bitmap), BITMAP_HEIGHT (bitmap), 1);
  unsigned int max_progress = BITMAP_HEIGHT (bitmap) * BITMAP_WIDTH (bitmap);

  for (row = 0; row < BITMAP_HEIGHT(bitmap); row++)
  {
    for (col = 0; col < BITMAP_WIDTH(bitmap); col++)
    {
      pixel_outline_type outline;
      unsigned branch, neighbours, marked_neighbours;

      if (notify_progress)
        notify_progress((at_real)(row * BITMAP_WIDTH(bitmap) + col) / ((at_real) max_progress * (at_real)3.0),
        progress_data);

      if ((GetColor(bitmap, row, col)==0xFF)
          || (GetColor(marked, row, col)==0))
        continue;
      neighbours=count_all_neighbours(bitmap, row, col);
      marked_neighbours=count_all_neighbours(marked, row, col);
      if((neighbours-marked_neighbours) > 0){
        if(neighbours > 2)
          neighbours=neighbours-marked_neighbours;
        else
          neighbours=1;
        /* find junctions which are neighbours */
        for(branch=0; branch < neighbours; branch++){
          LOG1("#%u: ", O_LIST_LENGTH(*outline_list));

          outline = find_one_centerline(bitmap, &marked, row, col);

          if (O_LENGTH(outline) > 1)
            append_pixel_outline(outline_list, outline);
          else
            LOG2("length 1 skipped %d %d\n", col, row);

          LOG1("%s)", (outline.open ? "open" : "closed"));
          LOG1(" [%u].\n", O_LENGTH(outline));
        }
      }
      MarkPixel(&marked, row, col);
    }
    if (test_cancel && test_cancel(testcancel_data))
      goto cleanup;
  }
  cleanup:
  free_bitmap(&marked);
  flush_log_output();
}

