/***************************************************************************
   drawwing.c  -  Convert wing bitmap to curves.
                             -------------------
    begin                : Mon Mar 11 2002
    copyright            : (C) 2002 by Adam Tofilski
    email                : rotofils@cyf-kr.edu.pl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xstd.h"
#include "math.h"
#include "thin-image.h"
#include "fit.h"
#include "output-eps.h"
#include "autotrace.h"
#include "centerline.h"
//#include "dw-output.h"
#include "drawwing.h"


#define SQUARE(x) ((x) * (x))
#define MOD(x) ((x) < 0 ? -x : x)

typedef struct{
  unsigned *data;
  unsigned length;
} unsigned_list_type;

void
del_short_otl(pixel_outline_list_type *outlines, unsigned otl_nr,
              unsigned too_short);

static unsigned_list_type
new_unsigned_list (void)
{
  unsigned_list_type list;

  list.data = NULL;
  list.length = 0;

  return list;
}

static void
free_unsigned_list (unsigned_list_type *list)
{
  if (list->length > 0){
    free (list->data);
    list->data = NULL;
    list->length = 0;
  }
}

static void
append_unsigned_list (unsigned_list_type *list, unsigned new_nr)
{
  list->length++;
  XREALLOC (list->data, list->length * sizeof(unsigned));
  list->data[list->length - 1] = new_nr;
}

/* Subtruct bgd from bitmap. */
unsigned
subtruct(at_bitmap_type *bmp, at_bitmap_type bgd)
{
  unsigned row, col;

  if (bmp->width != bgd.width)
    return 1;

  if (bmp->height != bgd.height)
    return 1;

  for (row = 0; row < bmp->height; row++)
    for (col = 0; col < bmp->width; col++)
    {
      unsigned bgd_color=GetColor(bgd, row, col);
      unsigned bmp_color=GetColor(*bmp, row, col);
      unsigned new_color;

      new_color = bmp_color + 255 - bgd_color;
      if(new_color > 255)
        new_color = 255;
      SetColor(bmp, row, col, new_color);
    }
  return 0;
}


unsigned
outline_threshold(at_bitmap_type bitmap)
{
  unsigned hist_tab[256];
  at_real average_tab[256];
  unsigned row, col, color, max_color=0;
  at_real min_average, max_average=0.0;
  unsigned range=4;          

  for (color = 0; color < 256; color++)
  {
    hist_tab[color]=0;
    average_tab[color]=0;
  }

  for (row = 0; row < bitmap.height; row++)
    for (col = 0; col < bitmap.width; col++)
      hist_tab[GetColor(bitmap, row, col)]++;
  for (color=range; color < 256-range; color++)
  {
    int rng = range;
	int i=-rng;
    average_tab[color]=0;
    while(i < rng)
    {
      average_tab[color]+=hist_tab[color+i];
      i++;
    }
    if (average_tab[color] > max_average)
    {
      max_average=average_tab[color];
      max_color=color;
    }
  }

  min_average=max_average;
  color=max_color - 1;
  while((average_tab[color] < min_average)&&(color > 0))
  {
    min_average=average_tab[color];
    color--;
  }
  return color;
}

/*
from autotrace fit.c modified
find a tangent to any point of and closed outline,
consider points in tangent_surround on both sides of a start_pixel.
*/
vector_type
tangent_to_pixels (pixel_outline_type outline, unsigned start_pixel,
                   unsigned tangent_surround)
{
  vector_type tangent = { 0.0, 0.0 };
  unsigned surround_pos;
  at_coord tangent_coord = outline.data[start_pixel];

  for (surround_pos = 0; surround_pos <= tangent_surround; surround_pos++){
    at_coord pixel_coord;
    unsigned pixel = start_pixel+surround_pos;
    if (pixel<outline.length)
      pixel_coord = outline.data[pixel];
    else
      pixel_coord = outline.data[pixel-outline.length];
    tangent = Vadd (tangent,
                    IPsubtract (pixel_coord, tangent_coord));
   }
  for (surround_pos = 0; surround_pos <= tangent_surround; surround_pos++){
    at_coord pixel_coord;
    int pixel = start_pixel-surround_pos;
    if (pixel<0)
      pixel_coord = outline.data[outline.length+pixel];
    else
      pixel_coord = outline.data[pixel];
    tangent = Vadd (tangent,
                    IPsubtract (tangent_coord, pixel_coord));//important argument order
   }

//  LOG2 ("(%.3f,%.3f).\n", tangent.dx, tangent.dy);
  return tangent;
}

/* retuns position of first pixel of outline with smallest x */
unsigned
GetLeftestPixel(pixel_outline_type outline)
{
  unsigned LeftestPixelX=outline.data[0].x;
  unsigned pixel, LeftestPixel=0;

  for(pixel=1; pixel<outline.length; pixel++)
    if(outline.data[pixel].x<LeftestPixelX){
      LeftestPixel=pixel;
      LeftestPixelX=outline.data[pixel].x;
    }
  return LeftestPixel;
}

/* Retuns position of last pixel of outline with greatest x coordinates.
   it should be upper end of the wing. */
unsigned
GetRightestPixel(pixel_outline_type outline)
{
  unsigned RightestPixelX=outline.data[0].x;
  unsigned pixel, RightestPixel=0;

  for(pixel=1; pixel<outline.length; pixel++)
    if(outline.data[pixel].x>=RightestPixelX){
      RightestPixel=pixel;
      RightestPixelX=outline.data[pixel].x;
    }
  return RightestPixel;
}

/* Find pixel which is most distant from line with start_pixel and end pixel. */
//powinno dzialac szybciej od getTip ale nie jest dokladne 
//wynik zalezy od poczatkowego polozenia skrzydla wzgledem pionu 
unsigned
GetDistantPixel(pixel_outline_type outline, unsigned start_pixel,
                unsigned end_pixel)
{
  double denominator, distance, max_distance=0.0;
  int difx, dify;
  unsigned pixel=start_pixel, distant_pixel=start_pixel;

  difx=outline.data[start_pixel].x-outline.data[end_pixel].x;
  dify=outline.data[start_pixel].y-outline.data[end_pixel].y;
  denominator=sqrt(SQUARE(difx)+SQUARE(dify));

  do
  {
    distance=MOD(dify*(outline.data[pixel].x-outline.data[end_pixel].x)
                 -difx*(outline.data[pixel].y-outline.data[end_pixel].x)
                 )/denominator;
    if(distance > max_distance)
    {
      max_distance=distance;
      distant_pixel=pixel;
    }
    pixel++;
    if(pixel > outline.length-1)
      pixel=0;
  }while(pixel != end_pixel);

  return distant_pixel;
}

/* Find pixel which is most distant from line with start_pixel and end pixel. */
unsigned
GetTip(pixel_outline_type outline, unsigned start_pixel,
                unsigned end_pixel)
{
  double distance, max_distance=0.0;
  unsigned pixel=start_pixel, distant_pixel=start_pixel;

  do
  {
    vector_type start_tip_vec;
    vector_type end_tip_vec;
	start_tip_vec=IPsubtract(outline.data[start_pixel],
                            outline.data[pixel]);
    end_tip_vec=IPsubtract(outline.data[end_pixel],
                            outline.data[pixel]);
    distance=magnitude(start_tip_vec)+magnitude(end_tip_vec);
    if(distance > max_distance)
    {
      max_distance=distance;
      distant_pixel=pixel;
    }
    pixel++;
    if(pixel > outline.length-1)
      pixel=0;
  }while(pixel != end_pixel);

  return distant_pixel;
}

void
GetWingEnds(pixel_outline_type outline, wing_type *wing)
{
  unsigned RightBorderX;

  RightBorderX=outline.data[wing->right].x;

  wing->LowerEnd = wing->bottom; //lower end is later then bottom
  while(outline.data[wing->LowerEnd].x < RightBorderX)
    wing->LowerEnd ++;

  wing->UpperEnd = wing->right;
}

wing_type
GetWing(pixel_outline_type WingOutline, unsigned surround)
{
  wing_type wing;
  at_real UpperAngle=0, LowerAngle=0;
  vector_type cross_vector, vertical_vector={0.0,-1.0}, horyzontal_vector={-1.0,0.0};

  wing.left=GetLeftestPixel(WingOutline);
  wing.right=GetRightestPixel(WingOutline);

  if(wing.left!=0)
    wing.top = wing.left-1;
  else
    wing.top = WingOutline.length-1;

  if(wing.left!=(WingOutline.length-1))
    wing.bottom = wing.left+1;
  else
    wing.bottom = 0;

  do{
    if(UpperAngle>LowerAngle){
      if(wing.top!=0)
        wing.top--;
      else
        wing.top=WingOutline.length-1;
    }else
      wing.bottom++;
    if (wing.bottom > WingOutline.length-1)
      FATAL1 ("GetWing bad wing.bottom value(%d)", wing.bottom);

    cross_vector=IPsubtract(WingOutline.data[wing.bottom],
                            WingOutline.data[wing.top]);
    UpperAngle=Vangle(cross_vector, //outline-surround add to options
                      tangent_to_pixels(WingOutline, wing.top, 2*surround));
    LowerAngle=Vangle(cross_vector,
                      tangent_to_pixels(WingOutline, wing.bottom, 2*surround));
//    LOG4("(%d,%d,%f,%f),",wing.top, wing.bottom, UpperAngle, LowerAngle);
  }while((UpperAngle<90)||(LowerAngle<90));

  if(Vangle(cross_vector, horyzontal_vector)<90.0)
    wing.rotation_angle=Vangle(cross_vector, vertical_vector);
  else
    wing.rotation_angle=360.0-Vangle(cross_vector, vertical_vector);

  wing.tip=GetTip(WingOutline, wing.top, wing.bottom);

  GetWingEnds(WingOutline, &wing);
  wing.bottom_coord.x = WingOutline.data[wing.bottom].x;
  wing.bottom_coord.y = WingOutline.data[wing.bottom].y;
  wing.tip_coord.x = WingOutline.data[wing.tip].x;
  wing.tip_coord.y = WingOutline.data[wing.tip].y;

//temporary inprecise assesment of wing width
  if(WingOutline.data[wing.LowerEnd].x >
     WingOutline.data[wing.UpperEnd].x)
    wing.length = WingOutline.data[wing.LowerEnd].x -
                  WingOutline.data[wing.tip].x;
  else
    wing.length = WingOutline.data[wing.UpperEnd].x -
                  WingOutline.data[wing.tip].x;

  wing.width = magnitude(cross_vector);

  LOG("\n");
  LOG3 ("wing.left %d(%d,%d)\n", wing.left, WingOutline.data[wing.left].x,
                                 WingOutline.data[wing.left].y);
  LOG3 ("wing.right %d(%d,%d)\n", wing.right, WingOutline.data[wing.right].x,
                                  WingOutline.data[wing.right].y);
  LOG3 ("wing.tip %d(%d,%d).\n", wing.tip, WingOutline.data[wing.tip].x,
                                 WingOutline.data[wing.tip].y);
  LOG3 ("wing.top %d(%d,%d).\n", wing.top, WingOutline.data[wing.top].x,
                                 WingOutline.data[wing.top].y);
  LOG3 ("wing.bottom %d(%d,%d).\n", wing.bottom, WingOutline.data[wing.bottom].x,
                                    WingOutline.data[wing.bottom].y);
  LOG3 ("wing.LowerEnd %d(%d,%d).\n", wing.LowerEnd,
                                      WingOutline.data[wing.LowerEnd].x,
                                      WingOutline.data[wing.LowerEnd].y);
  LOG3 ("wing.UpperEnd %d(%d,%d).\n", wing.UpperEnd,
                                      WingOutline.data[wing.UpperEnd].x,
                                      WingOutline.data[wing.UpperEnd].y);
  LOG1 ("UpperAngle %f\n", UpperAngle);
  LOG1 ("LowerAngle %f\n", LowerAngle);
  LOG1 ("wing width %f\n", wing.width);
  LOG1 ("wing length %f\n", wing.length);
  LOG1 ("wing.rotation_angle %f.\n", wing.rotation_angle);

  return wing;
}

void
AddBorder(at_bitmap_type *bitmap)
{
  unsigned int row, col;
  unsigned char color=0xFF;
  unsigned border=4;

  for (row = 0; row < border; row++)
    for (col = 0; col < bitmap->width; col++)
      SetColor(bitmap, row, col, color);
  for (row = 1; row < bitmap->height-1; row++){
    for (col = 0; col < border; col++)
      SetColor(bitmap, row, col, color);
    for (col = bitmap->width-border; col < bitmap->width; col++)
      SetColor(bitmap, row, col, color);
  }
  for (row = bitmap->height-border; row < bitmap->height; row++)
    for (col = 0; col < bitmap->width; col++)
      SetColor(bitmap, row, col, color);
}

//dodaj koncowki do zylek na prawym marginesie zeby dorze rozpoczac centerline
//wylaczone bo nie dziala dobrze
void
EndVeins(at_bitmap_type *bitmap)
{
  unsigned int row;
  unsigned char background=0xFF;
  unsigned char black=0x00;
  unsigned border=4;
  unsigned col1=bitmap->width-border;
  unsigned col2=bitmap->width-border-1;
  unsigned col3=bitmap->width-border-2;
  unsigned lower_edge;
  unsigned_list_type vein_list=new_unsigned_list();

  lower_edge=0;
  for (row = 0; row < bitmap->height; row++)
  {
    if(!GetColor(*bitmap,row,col1)&&
       !GetColor(*bitmap,row,col2)&&
       !GetColor(*bitmap,row,col3))
    {
      if(lower_edge==0)lower_edge=row;
    }else
      if(lower_edge>0)
      {
        append_unsigned_list(&vein_list,lower_edge+(row-lower_edge)/2);
        lower_edge=0;
      }
    SetColor(bitmap,row,col1,background);
    SetColor(bitmap,row,col2,background);
    SetColor(bitmap,row,col3,background);
  }
  for(row=0;row<vein_list.length;row++)
  {
    SetColor(bitmap,vein_list.data[row],col1,black);
    SetColor(bitmap,vein_list.data[row],col2,black);
    SetColor(bitmap,vein_list.data[row],col3,black);
  }
  free_unsigned_list(&vein_list);
}

void
ApplyThreshold(at_bitmap_type *bitmap, unsigned char threshold){
  unsigned int row, col;

  for (row = 0; row < bitmap->height; row++)
    for (col = 0; col < bitmap->width; col++)
      if(GetColor(*bitmap,row,col)<threshold)
        SetColor(bitmap, row, col, 0);
      else
        SetColor(bitmap, row, col, 0xFF);
}

void
CopyPixel(at_bitmap_type FromBmp, at_bitmap_type *ToBmp,
          unsigned row, unsigned col)
{
  SetColor(ToBmp, row, col, GetColor(FromBmp, row, col));
}


at_bitmap_type
CopyBmp(at_bitmap_type OldBmp)
{
  unsigned int row, col;
  at_bitmap_type NewBmp;

  NewBmp.width=OldBmp.width;
  NewBmp.height=OldBmp.height;
  NewBmp.np=OldBmp.np;
  XMALLOC (NewBmp.bitmap, NewBmp.width * NewBmp.height * sizeof (unsigned char));

  for (row = 0; row < OldBmp.height; row++)
    for (col = 0; col < OldBmp.width; col++)
      CopyPixel(OldBmp, &NewBmp, row, col);
  return NewBmp;
}

/**/
at_bitmap_type
MaskWing(at_bitmap_type OldBmp, pixel_outline_type outline, wing_type wing)
{
  unsigned pixelA, pixelB, minY, maxY, row, col;
  at_bitmap_type NewBmp;

  NewBmp.width=OldBmp.width;
  NewBmp.height=OldBmp.height;
  NewBmp.np=OldBmp.np;
  XMALLOC (NewBmp.bitmap, NewBmp.width * NewBmp.height * sizeof (unsigned char));

  for (row = 0; row < NewBmp.height; row++)
    for (col = 0; col < NewBmp.width; col++)
      SetColor(&NewBmp, row, col, 0xFF);

  for(pixelA=wing.left; pixelA<wing.right+1; pixelA++){
    minY=maxY=outline.data[pixelA].y;
    col=outline.data[pixelA].x;
    for(pixelB=0; pixelB<outline.length; pixelB++)//can be faster
      if(outline.data[pixelB].x==col)
        if(outline.data[pixelB].y>maxY)
          maxY=outline.data[pixelB].y;
    for(row=minY; row<=maxY; row++)
      CopyPixel(OldBmp, &NewBmp, NewBmp.height-row, col);
  }
  return NewBmp;
}

unsigned
fill_black(at_bitmap_type *bitmap, unsigned short row, unsigned short col)
{
  unsigned NewRow, NewCol;//, ShiftH, ShiftV;
  unsigned white = 0xFF;
  unsigned black = 0;
  unsigned dir = 1;
  unsigned count = 0;


  if(row<bitmap->height){
    NewRow=row+dir;
    if(GetColor(*bitmap, NewRow, col)==white){
      SetColor(bitmap, NewRow, col, black);
      count = fill_black(bitmap, NewRow, col);
      count++;
    }
    if(col<bitmap->width){
      NewCol=col+dir;
      if(GetColor(*bitmap, NewRow, NewCol)==white){
        SetColor(bitmap, NewRow, NewCol, black);
        count = fill_black(bitmap, NewRow, NewCol);
        count++;
      }
    }
    if(col>0){
      NewCol=col-dir;
      if(GetColor(*bitmap, NewRow, NewCol)==white){
        SetColor(bitmap, NewRow, NewCol, black);
        count = fill_black(bitmap, NewRow, NewCol);
        count++;
      }
    }
  }

  if(row>0){
    NewRow=row-dir;
    if(GetColor(*bitmap, NewRow, col)==white){
      SetColor(bitmap, NewRow, col, black);
      count = fill_black(bitmap, NewRow, col);
      count++;
    }
    if(col<bitmap->width){
      NewCol=col+dir;
      if(GetColor(*bitmap, NewRow, NewCol)==white){
        SetColor(bitmap, NewRow, NewCol, black);
        count = fill_black(bitmap, NewRow, NewCol);
        count++;
      }
    }
    if(col>0){
      NewCol=col-dir;
      if(GetColor(*bitmap, NewRow, NewCol)==white){
        SetColor(bitmap, NewRow, NewCol, black);
        count = fill_black(bitmap, NewRow, NewCol);
        count++;
      }
    }
  }

  if(col<bitmap->width){
    NewCol=col+dir;
    if(GetColor(*bitmap, row, NewCol)==white){
      SetColor(bitmap, row, NewCol, black);
      count = fill_black(bitmap, row, NewCol);
      count++;
    }
  }

  if(col>0){
    NewCol=col-dir;
    if(GetColor(*bitmap, row, NewCol)==white){
      SetColor(bitmap, row, NewCol, black);
      count = fill_black(bitmap, row, NewCol);
      count++;
    }
  }
  return count;
}

void
del_small_otl(at_bitmap_type *MaskedBmp, pixel_outline_list_type outlines,
              unsigned too_small)
{
  unsigned otl_nr;

  for(otl_nr=0; otl_nr<outlines.length; otl_nr++){
    pixel_outline_type otl=outlines.data[otl_nr];
    if((otl.length<too_small)
        &&(otl.color.r==255)){
      unsigned count;
      count = fill_black(MaskedBmp,
                         MaskedBmp->height - outlines.data[otl_nr].data[0].y,
                         outlines.data[otl_nr].data[0].x);
      LOG3("filling from point: %d, %d, %d\n",
            outlines.data[otl_nr].data[0].x,
            outlines.data[otl_nr].data[0].y, count);
    };
  }

}

void
smoothing1(at_bitmap_type *bitmap, unsigned size)
{
  unsigned row, col;
  int r,c, first=size * -1, last=size+1;
  at_bitmap_type TempBmp=CopyBmp(*bitmap);

  for (row = 0; row < bitmap->height; row++)
    for (col = 0; col < bitmap->width; col++)
      if(GetColor(TempBmp, row, col)==0xFF)
        for(r=first; r<last; r++)
          for(c=first; c<last; c++)
            SetColor(bitmap, row+r, col+c, 0xFF);

  TempBmp=CopyBmp(*bitmap);

  for (row = 0; row < bitmap->height; row++)
    for (col = 0; col < bitmap->width; col++)
      if(GetColor(TempBmp, row, col)==0)
        for(r=first; r<last; r++)
          for(c=first; c<last; c++)
            if((r<bitmap->height)&&(c<bitmap->width))
              SetColor(bitmap, row+r, col+c, 0);
}

void
smoothing2(at_bitmap_type *bitmap, unsigned size)
{
  unsigned row, col;
  int r,c; 
	const int first=size * -1;
	const int last=size+1;
  at_bitmap_type TempBmp=CopyBmp(*bitmap);

  for (row = 0; row < bitmap->height; row++){
	  for (col = 0; col < bitmap->width; col++){
		  if(GetColor(TempBmp, row, col)==0){
			  for(r=first; r<last; r++){
				  for(c=first; c<last; c++){
					  SetColor(bitmap, row+r, col+c, 0);
				  }
			  }
		  }
	  }
  }
  TempBmp=CopyBmp(*bitmap);

  for (row = size; row < (bitmap->height-size-1); row++){
	  for (col = size; col < (bitmap->width-size-1); col++){
		  if(GetColor(TempBmp, row, col)==0xFF){
			  for(r=first; r<last; r++){
				  for(c=first; c<last; c++){
					  SetColor(bitmap, row+r, col+c, 0xFF);
				  }
			  }
		  }
	  }
  }
}


/* Mark ends of veins at the border to prevent deletion. */
void
mark_end_veins(pixel_outline_list_type *outlines, unsigned border,
               unsigned too_short)
{
  unsigned otl_nr;

  for(otl_nr=0; otl_nr<outlines->length; otl_nr++)
    if((outlines->data[otl_nr].from_junction)
        || (outlines->data[otl_nr].to_junction))
    {
      if(outlines->data[otl_nr].data[0].x == border)
        {
          if(!outlines->data[otl_nr].from_junction)
            {
              XMALLOC(outlines->data[otl_nr].from_junction, sizeof(at_coord *));
              outlines->data[otl_nr].from_junction->x = border;
              outlines->data[otl_nr].from_junction->y
                = outlines->data[otl_nr].data[0].y;
            }
        }
      else
      if(outlines->data[otl_nr].data[outlines->data[otl_nr].length-1].x == border)
        {
          if(!outlines->data[otl_nr].to_junction)
            {
              XMALLOC(outlines->data[otl_nr].to_junction, sizeof(at_coord *));
              outlines->data[otl_nr].to_junction->x = border;
              outlines->data[otl_nr].to_junction->y
                = outlines->data[otl_nr].data[outlines->data[otl_nr].length-1].y;
            }
        }
    }
}

void
translate_splines(spline_list_array_type *splines, at_real delta_x,
                  at_real delta_y)
{
  unsigned list_nr, spline_nr;

  for (list_nr = 1; list_nr < SPLINE_LIST_ARRAY_LENGTH (*splines);list_nr++){
    spline_list_type *list = &SPLINE_LIST_ARRAY_ELT (*splines, list_nr);
    for (spline_nr = 0; spline_nr < SPLINE_LIST_LENGTH (*list);spline_nr++){
      spline_type *s = &SPLINE_LIST_ELT (*list, spline_nr);

      START_POINT (*s).x += delta_x;
      START_POINT (*s).y += delta_y;
      if (SPLINE_DEGREE (*s) == LINEARTYPE){
        END_POINT (*s).x += delta_x;
        END_POINT (*s).y += delta_y;
      }
      else{
        CONTROL1 (*s).x += delta_x;
        CONTROL1 (*s).y += delta_y;
        CONTROL2 (*s).x += delta_x;
        CONTROL2 (*s).y += delta_y;
        END_POINT (*s).x += delta_x;
        END_POINT (*s).y += delta_y;
      }
    }
  }
}

void
rotate_splines(spline_list_array_type *splines, double angle)
{
  unsigned list_nr, spline_nr;
  double sin_angle = sin(angle);
  double cos_angle = cos(angle);
  at_real x, y;

  for (list_nr = 1; list_nr < SPLINE_LIST_ARRAY_LENGTH (*splines);list_nr++){
    spline_list_type *list = &SPLINE_LIST_ARRAY_ELT (*splines, list_nr);
    for (spline_nr = 0; spline_nr < SPLINE_LIST_LENGTH (*list);spline_nr++){
      spline_type *s = &SPLINE_LIST_ELT (*list, spline_nr);

      x = START_POINT (*s).x;
      y = START_POINT (*s).y;
      START_POINT (*s).x = x*cos_angle-y*sin_angle;
      START_POINT (*s).y = y*cos_angle+x*sin_angle;
      if (SPLINE_DEGREE (*s) == LINEARTYPE){
        x = END_POINT (*s).x;
        y = END_POINT (*s).y;
        END_POINT (*s).x = x*cos_angle-y*sin_angle;
        END_POINT (*s).y = y*cos_angle+x*sin_angle;
      }
      else{
        x = CONTROL1 (*s).x;
        y = CONTROL1 (*s).y;
        CONTROL1 (*s).x = x*cos_angle-y*sin_angle;
        CONTROL1 (*s).y = y*cos_angle+x*sin_angle;
        x = CONTROL2 (*s).x;
        y = CONTROL2 (*s).y;
        CONTROL2 (*s).x = x*cos_angle-y*sin_angle;
        CONTROL2 (*s).y = y*cos_angle+x*sin_angle;
        x = END_POINT (*s).x;
        y = END_POINT (*s).y;
        END_POINT (*s).x = x*cos_angle-y*sin_angle;
        END_POINT (*s).y = y*cos_angle+x*sin_angle;
      }
    }
  }
}

void
scale_splines(spline_list_array_type *splines, at_real scale)
{
  unsigned list_nr, spline_nr;

  for (list_nr = 1; list_nr < SPLINE_LIST_ARRAY_LENGTH (*splines);list_nr++){
    spline_list_type *list = &SPLINE_LIST_ARRAY_ELT (*splines, list_nr);
    for (spline_nr = 0; spline_nr < SPLINE_LIST_LENGTH (*list);spline_nr++){
      spline_type *s = &SPLINE_LIST_ELT (*list, spline_nr);

      START_POINT (*s).x /= scale;
      START_POINT (*s).y /= scale;
      if (SPLINE_DEGREE (*s) == LINEARTYPE){
        END_POINT (*s).x /= scale;
        END_POINT (*s).y /= scale;
      }
      else{
        CONTROL1 (*s).x /= scale;
        CONTROL1 (*s).y /= scale;
        CONTROL2 (*s).x /= scale;
        CONTROL2 (*s).y /= scale;
        END_POINT (*s).x /= scale;
        END_POINT (*s).y /= scale;
      }
    }
  }
}

void
compare_wing(at_real_coord point, wing_type *wing)
{
  if(point.x < wing->real_left)
    wing->real_left = point.x;
  if(point.x > wing->real_right)
    wing->real_right = point.x;

  if(point.y < wing->lower)
    wing->lower = point.y;
  if(point.y > wing->upper)
    wing->upper = point.y;
}
void
get_splines_size(spline_list_array_type splines, wing_type *wing)
{
  unsigned list_nr, spline_nr;

  wing->real_left = splines.data[1].data[0].v[0].x;
  wing->real_right = 0;
  wing->lower = splines.data[1].data[0].v[0].y;
  wing->upper = 0;

  //start with 1 to omit border
  for (list_nr = 1; list_nr < SPLINE_LIST_ARRAY_LENGTH (splines);list_nr++){
    spline_list_type list = SPLINE_LIST_ARRAY_ELT (splines, list_nr);
    for (spline_nr = 0; spline_nr < SPLINE_LIST_LENGTH (list);spline_nr++){
      spline_type s = SPLINE_LIST_ELT (list, spline_nr);
      compare_wing(START_POINT (s), wing);
      compare_wing(END_POINT (s), wing);
    }
  }
  wing->big_width = wing->upper - wing->lower;
  wing->length = wing->real_right - wing->real_left;//usunac real* i zaokraglac
}

void
move_splines(spline_list_array_type *splines, wing_type *wing)
{
  at_real delta_x = -wing->bottom_coord.x;
  at_real delta_y = -wing->bottom_coord.y;
  double angle = wing->rotation_angle*3.14/180.0;//poprawic na dokladna wartosc obliczenia w geting
  at_real scale;

  translate_splines(splines,delta_x, delta_y);
  rotate_splines(splines, angle);
  get_splines_size(*splines, wing);
  delta_x = -wing->real_left;
  delta_y = -wing->lower;
  translate_splines(splines,delta_x, delta_y);
  if ((3*wing->width)<wing->big_width)  // error too small wing width TO DO send info to user
	wing->width = wing->big_width;      // do not produce huge images 
	
  scale = wing->width/100.0;
  scale_splines(splines, scale);

  splines->height = wing->big_width;
  splines->width = wing->length;
  splines->wing_width = wing->width;

}


at_splines_type *
dw_splines (at_bitmap_type * bitmap,
         at_fitting_opts_type * opts)
//         at_progress_func notify_progress,
//         at_address progress_data,
//         at_testcancel_func test_cancel,
//         at_address testcancel_data)
{
  at_splines_type * splines = NULL;
  pixel_outline_list_type pixels; //, temp_outlines;
  wing_type wing;
  at_bitmap_type WingOutline, MaskedBmp;
  color_type bg_color = { 0xff, 0xff, 0xff };

  if ((opts->outline == 0)||(opts->outline > 255))
    opts->outline = outline_threshold(*bitmap);

  LOG1("outline threshold = %d\n", opts->outline);
  printf("outline threshold = %d\n", opts->outline);

  WingOutline=CopyBmp(*bitmap);
  ApplyThreshold(&WingOutline, opts->outline);
  AddBorder(&WingOutline);


  pixels = find_outline_pixels(WingOutline, opts->bgColor,
                               NULL, NULL,
                               NULL, NULL);

  wing=GetWing(pixels.data[1], opts->tangent_surround);
  MaskedBmp=MaskWing(*bitmap, pixels.data[1], wing);

  if ((opts->venation == 0)||(opts->venation > 255))
    opts->venation = outline_threshold(MaskedBmp);

  LOG1("venation threshold = %d\n", opts->venation);
  printf("venation threshold = %d\n", opts->venation);

  ApplyThreshold(&MaskedBmp, opts->venation);
  smoothing2(&MaskedBmp, 2);

  thin_image (&MaskedBmp, opts->bgColor);

  if (opts->bgColor) bg_color = *opts->bgColor;

  find_centerline_pixels(&pixels, MaskedBmp, bg_color,
                         NULL, NULL,
                         NULL, NULL);

  mark_end_veins(&pixels, bitmap->width-4, opts->shortest_vein);

  find_short_otl(&pixels, opts->shortest_vein);


  XMALLOC(splines, sizeof(at_splines_type));
  *splines = fitted_splines (pixels, opts,
           bitmap->width,
           bitmap->height,
           NULL, NULL,
           NULL, NULL,
           wing);
  if(log_file){
    FILE *output_file;
    int llx, lly, urx, ury;

    llx = 0;
    lly = 0;
    urx = splines->width;
    ury = splines->height;

    output_file = xfopen("log.eps", "wb");
    output_eps_writer(output_file, " ", llx, lly, urx, ury, 300, *splines);
    fclose (output_file);
  }

  move_splines(splines, &wing);

  free_pixel_outline_list (&pixels);
  return splines;
}

const char *
dw_version (void)
{
  return "DrawWing version 0.12";
}

const char *
dw_home_site (void)
{
  return "http://kinga.cyf-kr.edu.pl/~rotofils/drawwing";
}



