/***************************************************************************
   pxl-outline.h  -  Find a list of outlines.
   Modified pxl-outline.h from Autotrace by Martin Weber.

                             -------------------
    begin                : Thu Sep 5 2002
    copyright            : (C) 2002 by Adam Tofilski
    email                : rotofils@cyf-kr.edu.pl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PXL_OUTLINE_H
#define PXL_OUTLINE_H

#include "autotrace.h"
#include "bitmap.h"
#include "color.h"

/* This is a list of contiguous points on the bitmap.  */
typedef struct
{
  at_coord *data;
  unsigned length;
  at_bool clockwise;
  color_type color;
  at_bool open;
  at_coord *from_junction; //AT
  at_coord *to_junction;
  unsigned nr; //AT
} pixel_outline_type;

/* The Nth coordinate in the list.  */
#define O_COORDINATE(p_o, n) ((p_o).data[n])

/* The length of the list.  */
#define O_LENGTH(p_o) ((p_o).length)

/* Whether the outline moves clockwise or counterclockwise.  */
#define O_CLOCKWISE(p_o) ((p_o).clockwise)

/* Since a pixel outline is cyclic, the index of the next coordinate
   after the last is the first, and the previous coordinate before the
   first is the last.  */
#define O_NEXT(p_o, n) (((n) + 1) % O_LENGTH (p_o))
#define O_PREV(p_o, n) ((n) == 0				\
                         ? O_LENGTH (p_o) - 1			\
                         : (n) - 1)

/* And the character turns into a list of such lists.  */
typedef struct
{
  pixel_outline_type *data;
  unsigned length;
} pixel_outline_list_type;

/* The Nth list in the list of lists.  */
#define O_LIST_OUTLINE(p_o_l, n) ((p_o_l).data[n])

/* The length of the list of lists.  */
#define O_LIST_LENGTH(p_o_l) ((p_o_l).length)

#ifdef _EXPORTING
/* Find all pixels on the outline in the character C.  */
extern pixel_outline_list_type __declspec(dllexport) __stdcall
find_outline_pixels (bitmap_type, color_type *bg_color,
		     at_progress_func notify_progress, address progress_data,
		     at_testcancel_func test_cancel, address testcancel_data);

//* Find all pixels on the center line of the character C.  */
//extern void __declspec(dllexport) __stdcall//AT
//find_centerline_pixels (pixel_outline_list_type *outline_list,
//			bitmap_type, color_type bg_color,
//			at_progress_func notify_progress, address progress_data,
//			at_testcancel_func test_cancel, address testcancel_data);

/* Free the memory in the list.  */
extern void __declspec(dllexport) __stdcall
free_pixel_outline_list (pixel_outline_list_type *);

#elif _IMPORTING
/* Find all pixels on the outline in the character C.  */
extern pixel_outline_list_type __declspec(dllimport) __stdcall
find_outline_pixels (bitmap_type, color_type *bg_color,
		     at_progress_func notify_progress, address progress_data,
		     at_testcancel_func test_cancel, address testcancel_data);

//* Find all pixels on the center line of the character C.  */
//extern void __declspec(dllimport) __stdcall//AT
//find_centerline_pixels (pixel_outline_list_type *outline_list,
//			bitmap_type, color_type bg_color,
//			at_progress_func notify_progress, address progress_data,
//			at_testcancel_func test_cancel, address testcancel_data);

/* Free the memory in the list.  */
extern void __declspec(dllimport) __stdcall
free_pixel_outline_list (pixel_outline_list_type *);

#else
/* Find all pixels on the outline in the character C.  */
extern pixel_outline_list_type
find_outline_pixels (bitmap_type, color_type *bg_color,
		     at_progress_func notify_progress, at_address progress_data,
		     at_testcancel_func test_cancel, at_address testcancel_data);

//* Find all pixels on the center line of the character C.  */
//extern void
//find_centerline_pixels (pixel_outline_list_type *outline_list,
//			bitmap_type, color_type bg_color,
//			at_progress_func notify_progress, at_address progress_data,
//			at_testcancel_func test_cancel, at_address testcancel_data);

/* Free the memory in the list.  */
extern void free_pixel_outline_list (pixel_outline_list_type *);

extern pixel_outline_type new_pixel_outline (void);//AT
extern void append_outline_pixel (pixel_outline_type *, at_coord);//AT
extern void
concat_pixel_outline(pixel_outline_type *o1, const pixel_outline_type *o2);
extern void
append_pixel_outline (pixel_outline_list_type *, pixel_outline_type);

#endif

#endif /* not PXL_OUTLINE_H */
